helpers={}
helpers.version = '1.05'

----------------------------------
--
-- Various functions
--
-- Author: Christof De Backere
--
----------------------------------

--dependencies above LM base
--lib: knxvalue
--functions:

--ToDo
--

--Known issues:

--log
--1.03 : added better support for high precision floats in table.save
--1.04: removed customsed alert: code needs to include debug condition
--1.05: lua compile: absolute path
--1.05: luacompile: first try luajit

require'knxvalue'

helpers.DEBUG = 0

--local _alert=alert or print
local function OLDalert(message,level)
  -- function to alert, considering a verbose level
  level=level or 1
  message=message or ""
  if level<=helpers.DEBUG then
    _alert(string.rep('_  ',level-1)
      ..string.gsub(message,'%%','/'))
--  	..string.gsub(message, "$(%b[])", function (x)
--             x = "return " .. string.sub(x, 2, -2)
--             local f = loadstring(x)
--             return tostring(f())
--           end))
  end
end

function trim(s)
--  return s:match'^%s*(.*%S)' or ''
 local n = s:find"%S"
 return n and s:match(".*%S", n) or ""
end


function utimediff(pt_sec, pt_usec)
   local ct_sec, ct_usec = os.microtime()
   return (ct_sec - pt_sec) + (ct_usec - pt_usec) / 1000000
end

function runScript(name, sync)
  local id, cmd
  id = db:getone('SELECT id FROM scripting WHERE name=?', name)
  id = tonumber(id)

  if id then
    cmd = string.format('lua /lib/genohm-scada/daemon/scripting-scheduled.lua %d', id)

    if not sync then
      cmd = cmd .. ' >/dev/null 2>/dev/null &'
    end

    os.execute(cmd)
  else
    error('[runScript] cannot find script: ' .. name)
  end
end

function serialTime2EIS(sT,dateortime)
--convert a lua serial timeStamp to the EIS dt.time and dt.date table format
  tT = os.date("*t",sT)
  if dateortime == "t" then
    return {day=tT.wday , hour=tT.hour , minute=tT.min , second=tT.sec }
  elseif dateortime == "d" then
    return {day=tT.day ,month=tT.month ,year=tT.year }
  else
    return {day=(tT.wday+5) % 7 + 1 , hour=tT.hour , minute=tT.min , second=tT.sec }, {day=tT.day ,month=tT.month ,year=tT.year }
  end
end


click=click or {}

click.DBLDELAY = 0.30 --default double click delay

function click.handle(event, DBLfunction, SNGfunction, dbldelay)
  --handles an event as a click and double-click
  --note: because the LM is rather slow, it is best to try to keep the single click action a direct action on the KNX bus, not needing the LM
  --current typical delay for a single click through the LM = 0.5-0.6s

  --minimize code before any action in order not to delay execution => alerts commented out
  nt, nut = os.microtime()
  local last = knxvalue.getprevious(event.dst)
  dbldelay = dbldelay or click.DBLDELAY
  --alert('1 '..utimediff(nt, nut))

  knxvalue.setlast(event.dst,{value = event.datahex, sec = nt, usec = nut})

  --alert('2 '..utimediff(nt, nut))

  if last then
--    alert("previous data: "..table.save(last),3)
    local deltaT = nt - last.sec
    deltaT = deltaT +(nut - last.usec)/1000000
--    alert ("check click "..event.dst..' '..deltaT..' '..last.sec..'+'..last.usec..' '..nt..'+'..nut..' '..utimediff(nt, nut),2)

    if dbldelay  > deltaT then
      if type(DBLfunction)=="function" then DBLfunction() end
      if helpers.DEBUG>=2 then alert ("doubleclick "..event.dst..' '..deltaT..' '..last.sec..'+'..last.usec..' '..nt..'+'..nut..' '..utimediff(nt, nut)) end
    else
      --already get knxobject to avoid to wait too long: time for knxobject in wait
--    singleClickObj = knxobject.get('address',"0/5/0")
--      alert("single click")
      local wt = dbldelay-utimediff(nt,nut)
      if wt>0 then os.sleep(wt) end
      local nlast = knxvalue.getprevious(event.dst)
      if nlast.sec==nt and nlast.usec==nut then
        --single click
        --knxvalue.set(true,"0/5/0") not very fast
        if type(SNGfunction)=="function" then SNGfunction() end
        if helpers.DEBUG>=2 then alert(" singleclick "..event.dst..' '..deltaT..' '..last.sec..'+'..last.usec..' '..nt..'+'..nut..' '..utimediff(nt, nut)) end
      else
        --'first of double click
--        alert("first click")
        if helpers.DEBUG>=2 then alert("first of doubleclick "..event.dst..' '..deltaT..' '..last.sec..'+'..last.usec..' '..nt..'+'..nut..' '..utimediff(nt, nut)) end
      end
    end
  else
    if helpers.DEBUG>=2 then alert("no previous click data") end
  end

  --alert('3 '..utimediff(nt, nut))
  --alert('4 '..utimediff(nt, nut))
end

----------------------------------------
--
-- File and Lib management
--
----------------------------------------

function getFile(fileName)
  local data = io.open(fileName,"r")
  local txt=""
  if data then
    txt = data:read("*all")
    --alert("reads data: "..txt,2) --can cause errors due to lua escape character in file content
    data:close()
  end

  return txt..""
end

function getLuaLib(libName)
  fileName='/usr/lib/lua/'..libName..'.lua'
  return getFile(fileName)
end

function setFile(fileName, text)
  local data = io.open(fileName,"w+")
  if data then
    data:write(text)
    data:flush()
    data:close()
    return true
  else
    return false
  end
end

function setFileC(fileName, libCode, keepScript)
  tmpFileName=fileName..'s'
  setFile(tmpFileName, libCode)
  os.execute('/usr/bin/luajit -b '..tmpFileName..' '..fileName)
  os.execute('/usr/bin/luac -o '..fileName..' '..tmpFileName)
--  alert("Writing compiled lua user lib '"..libName..".lua'")
  if keepScript or false then setFile(tmpFileName,"") end
--  return
end

function setLuaLib(libName, libCode)
  fileName='/usr/lib/lua/'..libName..'.lua'
  alert("Writing lua user lib '"..libName..".lua'")
  return setFile(fileName, libCode)
end

function setLuaLibC(libName, libCode, keepScript)
  fileName='/usr/lib/lua/'..libName..'.lua'
  alert("Writing compiled lua user lib '"..libName..".lua'")
  return   setFileC(fileName, libCode, keepScript)
end

userFunctionLib  = '/usr/share/lua/genohm-scada/userlib.lua'
initScript       = '/usr/share/lua/genohm-scada/initscript.lua'

-------------------------------------
--
--  table extension
--  table.save, table.load, table.count
--
-------------------------------------
--
--  Save Table to File/Stringtable
--  Load Table from File/Stringtable
--  v 0.94
--
--  Lua 5.1 compatible
--
--  Userdata and indices of these are not saved
--  Functions are saved via string.dump, so make sure it has no upvalues
--  References are saved
--  ----------------------------------------------------
--  table.save( table [, filename] )
--
--  Saves a table so it can be called via the table.load function again
--  table must a object of type 'table'
--  filename is optional, and may be a string representing a filename or true/1
--
--  table.save( table )
--    on success: returns a string representing the table (stringtable)
--    (uses a string as buffer, ideal for smaller tables)
--  table.save( table, true or 1 )
--    on success: returns a string representing the table (stringtable)
--    (uses io.tmpfile() as buffer, ideal for bigger tables)
--  table.save( table, "filename" )
--    on success: returns 1
--    (saves the table to file "filename")
--  on failure: returns as second argument an error msg
--  ----------------------------------------------------
--  table.load( filename or stringtable )
--
--  Loads a table that has been saved via the table.save function
--
--  on success: returns a previously saved table
--  on failure: returns as second argument an error msg
--  ----------------------------------------------------
--
--  chillcode, http://lua-users.org/wiki/SaveTableToFile
--  Licensed under the same terms as Lua itself.
do
  -- declare local variables
  --// exportstring( string )
  --// returns a "Lua" portable version of the string
  local function exportstring( s )
    s = string.format( "%q",s )
    -- to replace
    s = string.gsub( s,"\\\n","\\n" )
    s = string.gsub( s,"\r","\\r" )
    s = string.gsub( s,string.char(26),"\"..string.char(26)..\"" )
    return s
  end

  --// The Save Function
function table.save(  tbl,filename )
  local charS,charE = "   ","\n"
  local file,err
  -- create a pseudo file that writes to a string and return the string
  if not filename then
    file =  { write = function( self,newstr ) self.str = self.str..newstr end, str = "" }
    charS,charE = "",""
  -- write table to tmpfile
  elseif filename == true or filename == 1 then
    charS,charE,file = "","",io.tmpfile()
  -- write table to file
  -- use io.open here rather than io.output, since in windows when clicking on a file opened with io.output will create an error
  else
    file,err = io.open( filename, "w" )
    if err then return _,err end
  end
  -- initiate variables for save procedure
  local tables,lookup = { tbl },{ [tbl] = 1 }
  file:write( "return {"..charE )
  for idx,t in ipairs( tables ) do
    if filename and filename ~= true and filename ~= 1 then
      file:write( "-- Table: {"..idx.."}"..charE )
    end
    file:write( "{"..charE )
    local thandled = {}
    for i,v in ipairs( t ) do
      thandled[i] = true
      -- escape functions and userdata
      if type( v ) ~= "userdata" then
        -- only handle value
        if type( v ) == "table" then
          if not lookup[v] then
            table.insert( tables, v )
            lookup[v] = #tables
          end
          file:write( charS.."{"..lookup[v].."},"..charE )
        elseif type( v ) == "function" then
          file:write( charS.."loadstring("..exportstring(string.dump( v )).."),"..charE )
        else
          local value =  ( type( v ) == "string" and exportstring( v ) ) or tostring( v )
          file:write(  charS..value..","..charE )
        end
      end
    end
    for i,v in pairs( t ) do
      -- escape functions and userdata
      if (not thandled[i]) and type( v ) ~= "userdata" then
        -- handle index
        if type( i ) == "table" then
          if not lookup[i] then
            table.insert( tables,i )
            lookup[i] = #tables
          end
          file:write( charS.."[{"..lookup[i].."}]=" )
        else
          local index = ( type( i ) == "string" and "["..exportstring( i ).."]" ) or string.format( "[%d]",i )
          file:write( charS..index.."=" )
        end
        -- handle value
        if type( v ) == "table" then
          if not lookup[v] then
            table.insert( tables,v )
            lookup[v] = #tables
          end
          file:write( "{"..lookup[v].."},"..charE )
        elseif type( v ) == "function" then
          file:write( "loadstring("..exportstring(string.dump( v )).."),"..charE )
        elseif type( v ) == "number" then --extra float convert
          if tonumber(tostring(v))==v then
            file:write( tostring(v)..","..charE )
          else
            file:write( string.format('%f', v )..","..charE )
          end
        else
          local value =  ( type( v ) == "string" and exportstring( v ) ) or tostring( v )
          file:write( value..","..charE )
        end
      end
    end
    file:write( "},"..charE )
  end
  file:write( "}" )
  -- Return Values
  -- return stringtable from string
  if not filename then
    -- set marker for stringtable
    return file.str.."--|"
  -- return stringttable from file
  elseif filename == true or filename == 1 then
    file:seek ( "set" )
    -- no need to close file, it gets closed and removed automatically
    -- set marker for stringtable
    return file:read( "*a" ).."--|"
  -- close file and return 1
  else
    file:flush()
    file:close()
    return 1
  end
end

--// The Load Function
function table.load( sfile )
  -- catch marker for stringtable
  if string.sub( sfile,-3,-1 ) == "--|" then
    tables,err = loadstring( sfile )
  else
    tables,err = loadfile( sfile )
  end
  if err then return _,err
  end
  tables = tables()
  for idx = 1,#tables do
    local tolinkv,tolinki = {},{}
    for i,v in pairs( tables[idx] ) do
      if type( v ) == "table" and tables[ v[1] ] then
        table.insert( tolinkv,{ i,tables[ v[1] ] } )
      end
      if type( i ) == "table" and tables[ i[1] ] then
        table.insert( tolinki,{ i,tables[ i[1] ] } )
      end
    end
    -- link values, first due to possible changes of indices
    for _,v in ipairs( tolinkv ) do
      tables[idx][ v[1] ] = v[2]
    end
    -- link indices
    for _,v in ipairs( tolinki ) do
      tables[idx][ v[2] ],tables[idx][ v[1] ] =  tables[idx][ v[1] ],nil
    end
  end
  return tables[1]
end
-- close do
end

table.count = function(t)
  local count = 0
  for i,j in pairs(t) do
    count=count+1
  end
  return count
end

-------------------------------------
--
--  math extension
--  math.round
--
-------------------------------------

function math.round(number, decimals, method)
    decimals = decimals or 0
    local factor = 10 ^ decimals
    if (method == "ceil" or method == "floor") then return math[method](number * factor) / factor
    else return tonumber(("%."..decimals.."f"):format(number)) end
end

-------------------------------------
--
--  averages
--  average.update
--
-------------------------------------

averages = {}

function averages.update(serieName, lastValue, number)
  -- function to store a series of data. Takes a value to store, and returns the average of the last 'number' of arguments
  --feature to add: additional arg: persistent or no: if persistent, then use storage (not working currently) or /usr/.. area, if not persistent then use /tmp dir
  --local dataFileName='/tmp/avg-'..serieName
  local t,total,actualNumber=0, tonumber(lastValue), 1
  local data, newData={}, {total}

  if helpers.DEBUG>=2 then alert('averages.update for '..serieName..' with '..lastValue) end
  --data = table.load(dataFileName)  or {}
  data=storage.get('avg_'..serieName,{}) --or {}
  local ts,err
  if helpers.DEBUG>=2 then alert('table.load count = '..#data) end

  if #data>0 then

    for t = 1, number-1 do
      if data[t] ~= nil then
        actualNumber = t+1
        newData[t+1]=data[t]
        total = total+data[t]
      else
        break --no more values in table
      end
    end

  end

  ts, err = table.save(newData,1)
  err=err or ''
  if helpers.DEBUG>=2 then alert('tableStr = '..ts..' '..err) end
  --table.save(newData,dataFileName)
  storage.set('avg_'..serieName,newData)

  return total / actualNumber
end

-------------------------------------
--
--  os extension
--  os.capture
--
-------------------------------------

function os.capture(cmd, raw)
  local f = assert(io.popen(cmd, 'r'))
  local s = assert(f:read('*a'))
  f:close()
  if raw then return s end
  s = string.gsub(s, '^%s+', '')
  s = string.gsub(s, '%s+$', '')
  s = string.gsub(s, '[\n\r]+', ' ')
  return s
end
