module(..., package.seeall);

--knxvalue = {}

----------------------------------
--
-- knxvalue functionality
--
-- Author: Christof De Backere
--
----------------------------------

--dependencies above LM base
--lib      : helpers (table extension), but no require: helpers needs to be loaded first, which also load knxvalue (as 'part' of helpers)
--functions: /

--ToDo:
--

--Known issues:

--require'helpers'

version = '1.03.d2'

DEBUG = 0  --meaning: normally nothing to report

--local _alert=alert or print
local function OLDalert(message,level)
  -- function to alert, considering a verbose level
  level=level or 1
  message=message or ""
  if level<=DEBUG then
    _alert('KNXVALUE '..string.rep('_  ',level-1)
      ..string.gsub(message,'%%','/'))
--  	..string.gsub(message, "$(%b[])", function (x)
--             x = "return " .. string.sub(x, 2, -2)
--             local f = loadstring(x)
--             return tostring(f())
--           end))
  end
end

function get(address, aORn)
  --simplified way to get the data of a knxobject. Only works if datatype is configured correctly
  aORn = aORn or 'address'
  local obj = knxobject.get(aORn,address)
  if not obj  then
    aORn = 'name'
    obj = knxobject.get(aORn,address)
  end
  if not obj then
    if DEBUG>=2 then alert('KNXVALUE: get: Object '..address..' not found') end
    return nil
  else
    return obj.data
  end
end

function set(val, address, arg1, arg2)
  --simplified function to write data to a group address. Only works if datatype is configured correctly
  local aORn, datatype
  if DEBUG>=3 then alert('KNXVALUE: '..type(arg1)..' '..type(arg2)) end
  if type(arg1)=='string' then
    aORn=arg1
    datatype=arg2
  else
    aORn=arg2
    datatype=arg1
  end
  aORn = aORn or 'address'

  if aORn=='address' and datatype~=nil then
    return knxobject.write(address,val,datatype)
  else
    local obj = knxobject.get(aORn,address)
    if obj == nil then
      aORn = 'name'
      obj = knxobject.get(aORn,address)
    end
    if obj==nil then
      return nil
    else
      return obj:write(val)
      --obj = knxobject.get(aORn,address)
      --return obj.data
    end
  end
end

function status(l1, l2, l3, L1, L2, L3)
  --creates a string report of the status of the GA's : purpose for email commands
  --without arguments: all objects. Otherwise: limited range
  local addrStr, t1, t2, t3 = ""
  L1 = L1 or l1 or 15
  L2 = L2 or l2 or 7
  L3 = L3 or l3 or 255
  l1 = l1 or 0
  l2 = l2 or 0
  l3 = l3 or 0
  local retStr = 'Status report for group addresses in range '..l1..'-'..L1..'/'..l2..'-'..L2..'/'..l3..'-'..L3..'\n'..string.rep('-',57)..'\n\n'
retStr = retStr .. '   address name                             datatype            data         datahex\n\n'
  local obj

  local allObjs = allObj()
  if allObjs then
    if DEBUG>=3 then alert('KNXVALUE:   '..'status based on '..table.count(allObjs).." harvested objects from database:"..table.save(allObjs)) end
    for t1 = l1 , L1 do
      for t2 = l2 , L2 do
        for t3 = l3 , L3 do
          addrStr = t1..'/'..t2..'/'..t3
          if DEBUG>=4 then alert('KNXVALUE:    '..'testing '..addrStr) end
          if allObj[addrStr] then
            if DEBUG>=4 then alert('KNXVALUE:    '..addrStr..' found in allObj') end
            obj=knxobject.get('address',addrStr)
            if obj then
              rw=0
              if obj.datahex==nil then
                if DEBUG>=3 then alert('KNXVALUE:   '..'obj '..addrStr..' dataraw is nil') end
              else
                rw = "0x"..tostring(obj.datahex)
                if type(obj.data)=='table' then objdata=string.sub(table.save(obj.data),10,-7) else objdata=obj.data end
                retStr = retStr .. string.format("%10s %-30s %10s %15s %15s \n",addrStr,tostring(obj.name),tostring(obj.datatype), tostring(objdata),rw)
              end
            end
          end
        end
      end
    end
  else
    updateObjList(l1,l2,l3,L1,L2,L3)
    retStr = status(l1,l2,l3,L1,L2,L3)
 --[[   for t1 = l1 or 0, L1 or l1 or 15 do
      for t2 = l2 or 0, L2 or l2 or 7 do
        for t3 = l3 or 0, L3 or l3 or 255 do
          addrStr = t1..'/'..t2..'/'..t3
          obj=knxobject.get('address',addrStr)
          if obj ~= nil then
            retStr = retStr .. string.format("%10s %-30s %10s %15s %15s \n",addrStr,tostring(obj.name),tostring(obj.datatype), tostring(obj.data),obj.dataraw)
            --addrStr .. ', ' .. obj.name .. ', ' .. (obj.datatype or "") .. ', ' .. tostring(obj.data) .. ', ' .. obj.dataraw ..'\n')
          end
        end
      end
    end  --]]
  end
  return retStr
end

allObj = function()
  --return a stored table containing all addresses of existing objects
  --this table can be used to speed up looping through GA's, because knxobject() is very slow especially for non existing GA's

  local t = table.load('/tmp/knxvalue_allObj')
  if type(t) == "table" then return  t end
  return nil
end

function updateObjList(l1,l2,l3,L1,L2,L3)
  --updates the list of existing GA's by scanning them all : to be executed regularly when no activity is expected eg. dayly midnight
  if DEBUG>=2 then alert('KNXVALUE:  '.."allObj update started") end

  local  t1, t2, t3, addrStr, obj
  local  allObjs = {}

  for t1 = l1 or 0, L1 or l1 or 15 do --15 do
    for t2 =  l2 or 0, L2 or l2 or 7 do
      for t3 =  l3 or 0, L3 or l3 or 255 do --255
        addrStr = t1..'/'..t2..'/'..t3
        obj=knxobject.get('address',addrStr)
        if obj then allObjs[addrStr]=addrStr end
      end
    end
  end

  local fileName='/tmp/'..'knxvalue_allObj'
  table.save(allObjs,fileName)
  if DEBUG>=3 then alert('KNXVALUE:   '..table.save(allObjs)) end
  if DEBUG>=2 then alert('KNXVALUE:  '.."allObj list updated: "..#allObjs.." objects") end
  if DEBUG>=3 then alert('KNXVALUE:   '.."allObj read test: "..table.count(allObj()).." objects") end

end

storedEvents={} --end

function setlast(address,info)
  -- stores the information of an event for later use: when a new event on the same GA would not allow to get that information through knxobject/event
  address = string.gsub(address,'/','-')
  local fileName='/tmp/knxvalue_'..address
  local data = io.open(fileName,"w+")
  if not data then
	if DEBUG>=3 then alert('KNXVALUE:   '.."no file opened") end
  end

  data:write(table.save(info))
  data:flush()
  data:close()
  if DEBUG>=3 then alert('KNXVALUE:   '.."written on ".. address.." info: ".. table.save(info)) end
  --knxvalue.getprevious(address)
end

function getprevious(address)
  --return the information stored on an event. Is typically called when a new event occurs, hence the name of the function 'getprevious'
  address = string.gsub(address,'/','-')
  local fileName='/tmp/knxvalue_'..address
  local data = io.open(fileName,"r")

  if data then
    local txt = data:read()
    if not txt then txt = data:read() end --retry in case data is just being overwritten by next event

    local brol,se=string.find(txt,"--|")
    txt = string.sub(txt,1,se)
    local t = table.load(txt)
    if DEBUG>=3 then alert('KNXVALUE:   '.."data for "..address.." found :"..txt..' -> table: '..tostring(t)) end
    --if data then data.close() end
    return t
  else
    if DEBUG>=3 then alert('KNXVALUE:   '.."data for "..address.." not found.") end
    return nil
  end
end
