module(..., package.seeall);

-----------------------------
--
-- Scene functionality
--
-- Author: Christof De Backere
--
-----------------------------

--note: errors are not alerted but returned at the end of the function call

--dependencies above LM base
--lib      : knxvalue
--functions: /

--ToDo
--

--Known issues:

require"knxvalue"

version = '1.01'

function define(sceneName,...)
-- creates the table structure. If it exsists then ,no action
-- variable number of arguments: addresses or names of the objects in the scene
-- object is either defined by a string (status=write object) or a table with first the write and then the status object
-- optional last argument is addressOrName and defaults to 'address' (recognised values: address, name)
-- during definition, datatypes are determined based on info in database (required). If datatypes change: need to delete & redefine scene

  -- check if it already exsists
  local st=getSceneTable(sceneName)
  if st then return false, 'scene '..sceneName..' already defined' end

  -- check for last optional argument
  nObj = arg['n']
  if arg[arg['n']]=='address' or arg[arg['n']]=='name' then
    addressOrName = arg[arg['n']]
    nObj=nObj-1
  else addressOrName='address'
  end

  --if
  local wol, sol, dtol = {},{}, {} --write, status, datatype
  for i=1,nObj do
    if type(arg[i])=='table' then
      wol[i]=arg[i][1]
      sol[i]=arg[i][2]
      dtol[i]=knxobject.get(addressOrName,wol[i]).datatype or knxobject.get(addressOrName,sol[i]).datatype
    else
      wol[i]=arg[i]
      sol[i]=arg[i]
      dtol[i]=knxobject.get(addressOrName,wol[i]).datatype
    end
  end

  local st = {name = sceneName, writeObjects = wol, datatypes = dtol, statusObjects = sol, scenes={}, addressOrName = addressOrName, n=nObj}

  setSceneTable(sceneName,st)

  return true
end

function getSceneTable(sceneName)
-- private/internal function
  --remove special characters from scenename for storage
  sceneName = sceneName:gsub('[^-_%w]+', '_')
  return storage.get('scene_'..sceneName)
end

function setSceneTable(sceneName, sceneTable)
-- private/internal function
  --remove special characters from scenename for storage
  sceneName = sceneName:gsub('[^-_%w]+', '_')
  storage.set('scene_'..sceneName, sceneTable)
end

function reset(sceneName)
-- function to reset the stored scenes values. Keeps the object definitions
  local st=getSceneTable(sceneName)
  if not st then return false, 'scene '..sceneName..' does not exist' end

  st.scenes={}

  --rescan for updated datatypes: to be added

  setSceneTable(sceneName,st)
  return true
end

function delete(sceneName)
-- function to delete all information of the scene
  local st=getSceneTable(sceneName)
  if not st then return false, 'scene '..sceneName..' does not exist' end

  setSceneTable(sceneName, nil)
  return true
end

function store(sceneName,sceneNumber)
-- function to read a scene status from the bus and store it as a sceneNumber
  if sceneNumber>=0 and sceneNumber<=63 then
    local st=getSceneTable(sceneName)
    if not st then return false, 'scene '..sceneName..' does not exist' end
    local vt={}
    for i,v in ipairs(st.statusObjects) do
      vt[i] = knxvalue.get(v,st.addressOrName)
    end
    st.scenes[tostring(sceneNumber)]=vt
    setSceneTable(sceneName,st)
  else
    --raise error
    return false, 'impossible sceneNumber'
  end
  return true
end

function storeValues(sceneName,sceneNumber,...)
-- function to define the values to be stored as a sceneNumber by code
-- variable arguments: number required equal to number of objects
  if sceneNumber>=0 and sceneNumber<=63 then
    local st=getSceneTable(sceneName)
    if not st then return false, 'scene '..sceneName..' does not exist' end
    if arg['n']~= st.n then return false, 'wrong number of arguments' end
    local vt={}
    for i,v in ipairs(st.statusObjects) do
      vt[i] = arg[i]
    end
    st.scenes[tostring(sceneNumber)]=vt
    setSceneTable(sceneName,st)
  else
    --raise error
    return false, 'impossible sceneNumber'
  end
  return true
end

function recall(sceneName,sceneNumber)
-- function to retrieve a sceneNumber and write it to the bus
  local st=getSceneTable(sceneName)

  if st then
    if st.scenes[tostring(sceneNumber)]==nil then return false, 'sceneNumber '..sceneNumber..' not defined for scene '..sceneName end
  else
    return false, 'scene '..sceneName..' does not exist'
  end

  for i,v in ipairs(st.writeObjects) do
    knxvalue.set(st.scenes[tostring(sceneNumber)][i],v,st.addressOrName,st.datatypes[i])
  end
  return true
end
