module(..., package.seeall);

----------------------------------------
--
-- trend functionality
--
-- Author: Christof De Backere
--
----------------------------------

--dependencies above LM base
--lib: /
--functions: / (all alerts commented out)

--ToDo: ade alert, make module
--

--Known issues:

version = '1.01'

DEBUG = 1

---storage naming issue workaround
storage._set = storage.set
storage._get = storage.get

local storage.set = function(key, value)
  key = key:gsub('[^-_%w]+', '_')
  return storage._set(key, value)
end

local storage.get = function(key, default)
  key = key:gsub('[^-_%w]+', '_')
  return storage._get(key, default)
end

function storeOLD(event,period)
--stores a new datapoint to the trend series and returns the linear regresion b(slope) and S(error)
--event is the event triggered by a object update. event can also be any table where the field 'dst' contains the object's address
--period = time range for trand analysis in seconds
  local olddata = storage.get('trend_'..event.dst) or {}

  local last=knxobject.get('address',event.dst)
  last.updatetime = tonumber(last.updatetime)
  last.daytime=last.updatetime/60/60/24
  --convert data as bool to int
  if type(last.data) == type(true) then last.data = (last.data and 1 ) or 0 end
--print(last.updatetime, last.data)

  local newdata={last}
  local cnt=2
  for i,v in ipairs(olddata) do
    --alert(type(v.updatetime)..' '..type(newdata[1].updatetime)..' '..type(period)..' '..v.updatetime..' '..newdata[1].updatetime..' '..period,0)
    if v.updatetime >= newdata[1].updatetime-period then
      newdata[cnt]=v
      cnt=cnt+1
    end
  end
  if #olddata>=cnt-1 then newdata[cnt]=olddata[cnt-1] end

  storage.set('trend_'..event.dst, newdata)

  local a, b, S = linReg(newdata,"daytime","data")
  return  a, b, S, a+b*last.daytime
end

function store(event,period)
--stores a new datapoint to the trend series and returns the linear regresion b(slope) and S(error)
--event is the event triggered by a object update. event can also be any table where the field 'dst' contains the object's address
--period = time range for trand analysis in seconds
  local olddata = storage.get('trend_'..event.dst) or {}

  local last=knxobject.get('address',event.dst)
  last.updatetime = tonumber(last.updatetime)
  last.daytime=last.updatetime/(60*60*24)
  --convert data as bool to int
  --if type(last.data) == type(true) then last.data = (last.data and 1 ) or 0 end

  secs = last.updatetime-last.daytime*3600*24
  --alert(last.updatetime.." "..string.format("%.8f  %.8f",last.daytime, last.updatetime/(60*60*24)).." "..secs)

  local newdata={last}
  local cnt=2
  for i,v in ipairs(olddata) do
    --alert(type(v.updatetime)..' '..type(newdata[1].updatetime)..' '..type(period)..' '..v.updatetime..' '..newdata[1].updatetime..' '..period,0)
    if v.updatetime >= newdata[1].updatetime-period then
      newdata[cnt]=v
      cnt=cnt+1
    end
  end
  if #olddata>=cnt-1 then newdata[cnt]=olddata[cnt-1] end

  storage.set('trend_'..event.dst, newdata)

  local a, b, S = linReg(newdata,"updatetime","data")
  return  a, b*(24*3600), S, a+b*last.updatetime
end

function linReg(dataref, x, y)
--returns the linear regression values of a data series: b(slope) and S(error)
--dataref is either a table where each datapoint is another table or is an object address, corresponding  with the event address used earlier with trend.store
--x and y determine which fields from the datapoints are to be used as x and y values
  local ld
  if type(dataref)=='table' then
    ld=dataref
  else
    ld=storage.get('trend_'..dataref)
  end
--log(ld)

  --no absolut x-scale needed=> substract minimum to avoid rounding errors with squares (eg if x = time in seconds)
  local minx = math.huge
  for i,v in ipairs(ld) do
    if minx>v[x] then minx=v[x] end
  end

  local cx, cy, cxx, cxy, cyy=0,0,0,0,0
  for i,v in ipairs(ld) do
    if type(v[y]) == type(true) then v[y] = (v[y] and 1 ) or 0 end
    cx  = cx  + (v[x]-minx)
    cy  = cy  + v[y]
    cxx = cxx + (v[x]-minx)*(v[x]-minx)
    cxy = cxy + (v[x]-minx)*v[y]
    cyy = cyy + v[y]*v[y]
  end
  local n=#ld
  --print(n,cx,cy,cxx,cxy,cyy,(n*cxx-cx*cx))

  local a, b, S = 0, 0, 0
  if n>1 then
    b=(n*cxy - cx*cy)/(n*cxx-cx*cx)
    a=(cy-b*cx)/n-b*minx
  end
  if n>2 then S=(n*cyy - cy*cy - b*b*(n*cxx-cx*cx))/n/(n-2) end

  return a, b, S
end

function avg(dataref, y, x)
  local data
  if type(dataref)=='table' then
    data=dataref
  else
    data=storage.get('trend_'..dataref) or {}
  end
log(data)

  local now=os.time()
  local minx, maxx, miny, maxy = math.huge, -math.huge, math.huge, -math.huge
  for i,v in ipairs(data) do
    if type(v[y]) == type(true) then v[y] = (v[y] and 1 ) or 0 end
    if minx>v[x] then minx=v[x] end
    if maxx<v[x] then maxx=v[x] end
    if miny>v[y] then miny=v[y] end
    if maxy<v[y] then maxy=v[y] end
  end
--log(minx, os.time(), #data)

  local wsumy, sumy = 0, 0
  for i,v in ipairs(data) do
    sumy = sumy + v[y]
    if i > 1 then
      wsumy = wsumy + (data[i-1][x]-v[x])*v[y]
--log(i, data[i-1][x],v[x],v[y],avg)
    else
      wsumy = wsumy + (now-v[x])*v[y]
--log(i, os.time(),v[x],v[y],avg)
    end
  end
--log(avg)

  return wsumy/(now-minx), sumy/#data
end

--averages = {}
--function averages.Update(serieName, lastValue, number)
function simpleAvgUpdate(serieName, lastValue, number)
  -- function to store a series of data. Takes a value to store, and returns the average of the last 'number' of arguments
  --feature to add: additional arg: persistent or no: if persistent, then use storage (not working currently) or /usr/.. area, if not persistent then use /tmp dir
  --local dataFileName='/tmp/avg-'..serieName
  local t,total,actualNumber=0, tonumber(lastValue), 1
  local data, newData={}, {total}

  --alert('averages.Update for '..serieName..' with '..lastValue,3)
  data=storage.get('trend-simple_'..serieName,{}) --or {}
  local ts,err

  if #data>0 then

    for t = 1, number-1 do
      if data[t] ~= nil then
        actualNumber = t+1
        newData[t+1]=data[t]
        total = total+data[t]
      else
        break --no more values in table
      end
    end

  end

  storage.set('trend-simple_'..serieName,newData)

  return total / actualNumber
end

